#include "mevaluator.h"
#include <QDebug>

#define FSIZE 100


MEvaluator::MEvaluator(QObject *parent)
    : QTcpServer(parent)
{
    this->parent = parent;
}

void MEvaluator::incomingConnection(int socketDescriptor)
{
    sck.setSocketDescriptor(socketDescriptor);
    getFrames();
}

int MEvaluator::getFrames()
{
    QQueue<int> last_movements;
    QQueue<int> medianlist;

    for (int i=0; i<10; i++)
    {
        last_movements.enqueue(0);
        medianlist.enqueue(0);
    }

    CvCapture *input_video = cvCaptureFromFile("C:\\sample2.avi");

    if (input_video == NULL)
    {
        qDebug()<<"Error: Can't open video.\n";
        return -1;
    }

    cvQueryFrame( input_video );
    CvSize frame_size;
    int fps = ( int )cvGetCaptureProperty( input_video, CV_CAP_PROP_FPS );

    frame_size.height = (int) cvGetCaptureProperty( input_video, CV_CAP_PROP_FRAME_HEIGHT );
    frame_size.width = (int) cvGetCaptureProperty( input_video, CV_CAP_PROP_FRAME_WIDTH );

    long number_of_frames;

    cvSetCaptureProperty( input_video, CV_CAP_PROP_POS_AVI_RATIO, 1. );
    number_of_frames = (int) cvGetCaptureProperty( input_video, CV_CAP_PROP_POS_FRAMES );
    cvSetCaptureProperty( input_video, CV_CAP_PROP_POS_FRAMES, 0. );
    //cvNamedWindow("Optical Flow", CV_WINDOW_AUTOSIZE);

    long current_frame = 0;
    int accum = 0, counter = 0;
    double ltotal = 0;

    while(true)
    {
        double hyp_total=0;
        static IplImage *frame = NULL, *frame_yuv = NULL, *frame1 = NULL, *frame1_1C = NULL, *frame2_1C = NULL, *eig_image = NULL, *temp_image = NULL, *pyramid1 = NULL, *pyramid2 = NULL;

        cvSetCaptureProperty( input_video, CV_CAP_PROP_POS_FRAMES, current_frame );
        frame = cvQueryFrame( input_video );

        if (frame == NULL)
        {
            qDebug() << "Error: Hmm. The end came sooner than we thought";
            return -1;
        }

        allocateOnDemand( &frame1_1C, frame_size, IPL_DEPTH_8U, 1 );
        allocateOnDemand( &frame_yuv, frame_size, IPL_DEPTH_8U, 1 );
        cvConvertImage(frame, frame1_1C, CV_CVTIMG_FLIP);
        allocateOnDemand( &frame1, frame_size, IPL_DEPTH_8U, 3 );
        cvConvertImage(frame, frame1, CV_CVTIMG_FLIP);

        frame = cvQueryFrame( input_video );
        if (frame == NULL)
        {
            qDebug() << "Error: Hmm. The end came sooner than we thought";
            return -1;
        }

        allocateOnDemand( &frame2_1C, frame_size, IPL_DEPTH_8U, 1 );
        cvConvertImage(frame, frame2_1C, CV_CVTIMG_FLIP);
        allocateOnDemand( &eig_image, frame_size, IPL_DEPTH_32F, 1 );
        allocateOnDemand( &temp_image, frame_size, IPL_DEPTH_32F, 1 );

        CvPoint2D32f frame1_features[FSIZE];

        int number_of_features;
        number_of_features = FSIZE;
        cvGoodFeaturesToTrack(frame1_1C, eig_image, temp_image, frame1_features, &number_of_features, .01, .01, NULL);

        CvPoint2D32f frame2_features[FSIZE];

        char optical_flow_found_feature[FSIZE];

        float optical_flow_feature_error[FSIZE];

        CvSize optical_flow_window = cvSize(3,3);

        CvTermCriteria optical_flow_termination_criteria = cvTermCriteria( CV_TERMCRIT_ITER | CV_TERMCRIT_EPS, 20, .3 );

        allocateOnDemand( &pyramid1, frame_size, IPL_DEPTH_8U, 1 );
        allocateOnDemand( &pyramid2, frame_size, IPL_DEPTH_8U, 1 );

        cvCalcOpticalFlowPyrLK(frame1_1C, frame2_1C, pyramid1, pyramid2, frame1_features, frame2_features, number_of_features, optical_flow_window, 5, optical_flow_found_feature, optical_flow_feature_error, optical_flow_termination_criteria, 0 );

        double * ists = cvAvg(frame1).val;
        ltotal += ists[0]*-0.14713 - 0.28886 *ists[1] + 0.436 *ists[2];

        /*
        for(int i = 0; i < number_of_features; i++)
        {
            if ( optical_flow_found_feature[i] == 0 )
                continue;

            int line_thickness;
            line_thickness = 1;

            CvScalar line_color;
            line_color = CV_RGB(255,0,0);

            CvPoint p,q;
            p.x = (int) frame1_features[i].x;
            p.y = (int) frame1_features[i].y;
            q.x = (int) frame2_features[i].x;
            q.y = (int) frame2_features[i].y;

            double angle;
            angle = atan2( (double) p.y - q.y, (double) p.x - q.x );
            double hypotenuse;
            hypotenuse = sqrt( square(p.y - q.y) + square(p.x - q.x) );

            hyp_total += hypotenuse;

            q.x = (int) (p.x - 3 * hypotenuse * cos(angle));
            q.y = (int) (p.y - 3 * hypotenuse * sin(angle));

            cvLine( frame1, p, q, line_color, line_thickness, CV_AA, 0 );

            p.x = (int) (q.x + 9 * cos(angle + pi / 4));
            p.y = (int) (q.y + 9 * sin(angle + pi / 4));
            cvLine( frame1, p, q, line_color, line_thickness, CV_AA, 0 );
            p.x = (int) (q.x + 9 * cos(angle - pi / 4));
            p.y = (int) (q.y + 9 * sin(angle - pi / 4));
            cvLine( frame1, p, q, line_color, line_thickness, CV_AA, 0 );
        }
        */
        cvConvertImage(frame1, frame1, CV_CVTIMG_FLIP);
        //cvShowImage("Optical Flow", frame1);

        int key_pressed;
        key_pressed = cvWaitKey(1000/fps);

        current_frame++;

        if (current_frame < 0)
            current_frame = 0;

        if (current_frame >= number_of_frames - 1)
            current_frame = number_of_frames - 2;

        last_movements.dequeue();
        last_movements.enqueue(hyp_total);

        int value = summer(last_movements);

        accum += value;
        counter++;

        if (counter==9)
        {
            counter = 0;
            evaluateFrames(findmedian(medianlist, accum), ltotal/10);
            accum = 0;
            ltotal = 0;
        }
    }
}

void MEvaluator::evaluateFrames(int movement, double lightinfo)
{
    char light;

    if (lightinfo > 2.40)
    {
        light = '0';
    }
    else if (lightinfo > 0.4)
    {
        light = '1';
    }
    else
    {
        light = '2';
    }

    qDebug() << "Light value: " << light<<"\n";
    char message[4];
    message[0] = movement + '0';
    message[1] = '&';
    message[2] = light;
    message[3] = '\0';

    sck.write(message);
}

void MEvaluator::allocateOnDemand( IplImage **img, CvSize size, int depth, int channels)
{
    if ( *img != NULL ) return;
        *img = cvCreateImage( size, depth, channels );

    if ( *img == NULL )
    {
        fprintf(stderr, "Error: Couldn't allocate image. Out of memory?\n");
        exit(-1);
    }
}

double MEvaluator::square(int a)
{
    return a * a;
}

int MEvaluator::summer(QQueue<int> &list)
{
    int sum = 0, tmp;
    QQueue<int> tmp_list;

    while (!list.isEmpty())
    {
        tmp = list.dequeue();
        sum += tmp;
        tmp_list.enqueue(tmp);
    }

    while (!tmp_list.isEmpty())
    {
        list.enqueue(tmp_list.dequeue());
    }

    return sum;
}

int MEvaluator::findmedian(QQueue<int> &list,int accum)
{
    int last_val, tmp;
    QQueue<int> tmp_list;
    int vals[4] = {0,0,0,0};

    if (accum<50000)
    {
        last_val = 0;
    }
    else if (accum<120000)
    {
        last_val = 1;
    }
    else if (accum<200000)
    {
        last_val = 2;
    }
    else
    {
        last_val = 3;
    }

    list.dequeue();

    while (!list.isEmpty())
    {
        tmp = list.dequeue();
        vals[tmp]++;
        tmp_list.enqueue(tmp);
    }

    vals[last_val]++;

    while (!tmp_list.isEmpty())
    {
        list.enqueue(tmp_list.dequeue());
    }

    list.enqueue(last_val);

    last_val = 0;

    for (int i=0; i<4; i++)
    {
        if (vals[i]>last_val)
        {
            last_val = vals[i];
            tmp = i;
        }
    }

    switch (tmp)
    {
    case 0:
        qDebug() << "Movement: Standing!";
        break;
    case 1:
        qDebug() << "Movement: Walking!";
        break;
    case 2:
        qDebug()<< "Movement: Running!";
        break;
    case 3:
        qDebug() << "Movement: Running Desperately!";
        break;
    }

    return tmp;
}

void MEvaluator::closeSocket()
{
    sck.disconnectFromHost();
    sck.waitForDisconnected();
}

